/* IOCTL, FINDFILE implementation (int 21h) */

#include <dirent.h>
#include <ctype.h>

extern int last_drive;


static __inline__ int int21_emu_ioctl ( void )
{
  switch (AL) {
    case 0  : { /* get device attrs */
                int ret=16384+8129;
                if (BX==0) ret|=1;
                if (BX==1) ret|=2;
                CLC;
                AL=ret;
                return 1;
              }
    
    case 1 : { /* set device attrs */
               CLC; /* ignore */
               return 1;
             }
    
    case 2 : { /* read from device */
               char *p=DOS_PB(DS,DX);
               int ret=read(BX,p,CX);
               if (ret<0) {
                 AX=DERRNO(errno_u2d(errno));
                 STC;
               } else {
                 AX=ret;
                 CLC;
               }
               return 1;
             }
    
    case 3 : { /* write to device */
               char *p=DOS_PB(DS,DX);
               int ret=write(BX,p,CX);
               if (ret<0) {
                 AX=DERRNO(errno_u2d(errno));
                 STC;
               } else {
                 AX=ret;
                 CLC;
               }
               return 1;
             }
						 
		case 8 : AX=1;      /* check for removable block device, TODO !! */
						 CLC;
						 return 1;

		case 0x0E : AL=0;         /* get logical drive mapping */
								CLC;
								return 1;

		case 0x0F : STC;                 /* set logical drive mapping */
								AX=DERRNO(DNOSYS);
								return 1;
            
    default :
#ifdef CONFIG_DEBUG_UNIMPLEMENTED_FUNC
      if (stddebug) fprintf(stddebug,"INT21: unimplemented IOCTL function (44h, %02Xh) at %04X:%04X.\n",AL,oldcs,oldip);
#endif
      return 1;
  }
}





struct dta_sys_vars_st {
  DIR *dir;
  char *path;
  char *mask;
  int attr;
  int root;
};

#define DTA ((struct dta_sys_vars_st*)(dta))


static __inline__ int check_dos_fnamechar ( char c , char d )
{
  if (c>='A'&&c<='Z') c+=32;
  if (d>='A'&&d<='Z') d+=32;
  return c==d;
}


static word attr_u2d ( mode_t m )
{
  word ret=0;
  if (m&S_IWUSR) ret|=1;
  if (S_ISDIR(m)) ret|=16;
  return ret;
}


static void ftime_u2d ( time_t t , word *d_date , word *d_time )
{
  struct tm *unx=localtime(&t);
  *d_date=((unx->tm_year-80)<<9) | ((1+unx->tm_mon)<<5) | unx->tm_mday;
  *d_time=(unx->tm_hour<<11) | (unx->tm_min<<5) | unx->tm_sec>>1;
}


static int check_dos_subfmask ( char *name , char *mask , int max )
{
  while (max--&&*mask&&*name&&(*mask=='?'||check_dos_fnamechar(*mask,*name))) {
    if (isupper(*name)) return 0;  /* file names contain capital letter are unavaiable for DOSRUN */
    mask++;
    name++;
  }
  return *mask=='*'||max==0||(*mask==0&&*name==0);
}


static int check_dos_fmask ( char *name , char *dta )
{
  char *n=dosrun_strdup(name);
  char *m=dosrun_strdup(DTA->mask);
  char *nd=strrchr(n,'.');
  char *md=strrchr(m,'.');
  int ok;
  if (strlen(name)>CONFIG_MAX_DOS_FN_SIZE) {
    if (stddebug) fprintf(stddebug,"FINDFILE: Dropping \"%s\" for its length.\n",name);
    free(n);
    free(m);
    return 0;
  }
  /* file names starting with `.' are invalid except for names `.' and `..', except if they're in DOS drive's root directory */
  if (!(*name!='.'||(*name=='.'&&name[1]=='\0'&&!DTA->root)||(*name=='.'&&name[1]=='.'&&name[2]=='\0'&&!DTA->root))) {
    if (stddebug) fprintf(stddebug,"FINDFILE: Dropping \"%s\" for beginner period/root directory.\n",name);
    free(n);
    free(m);
    return 0;
  }
  if (stddebug) fprintf(stddebug,"FINDFILE: check_dos_fmask %s %s ",name,DTA->mask);
  if (md && !nd && md[1]=='*') ok=1; else
  if ((md&&!nd)||(!md&&nd)) ok=0; else {
    if (nd) {
      *nd='\0';
      *md='\0';
      ok=check_dos_subfmask(n,m,256)&&check_dos_subfmask(nd+1,md+1,256);
    } else
      ok=check_dos_subfmask(n,m,256);
  }
  free(n);
  free(m);
  if (stddebug) fprintf(stddebug,"%d\n",ok);
  return ok;
}



static __inline__ int check_dos_amask ( mode_t m , int mask )
{
  if (stddebug) fprintf(stddebug,"FINDFILE: check_dos_amask: mask=%d mode=%d\n",mask,m);
  if ((mask&16)==0&&S_ISDIR(m)) return 0;
  return 1;
}



__inline__ static void fname_u2d ( char *unx , char *dos )
{
  while (*unx) *(dos++)=toupper(*(unx++));
  *dos='\0';
}



static int em_findfile ( char *dta )
{
  int ok=0;
  struct dirent *ent;
  struct stat st;
  do {
    ent=readdir(DTA->dir);
    if (ent) {
      if (check_dos_fmask(ent->d_name,dta)) {
        strcpy(path,DTA->path);
	strcat(path,"/");
	strcat(path,ent->d_name);
	stat(path,&st);
	ok=check_dos_amask(st.st_mode,DTA->attr);
      }
    }
  } while (ent&&!ok);
  if (ok&&ent) {
    fname_u2d(ent->d_name,dta+30);
    *(byte*)(dta+21)=attr_u2d(st.st_mode);
    *(unsigned int *)(dta+26)=st.st_size;
    ftime_u2d(st.st_mtime,(word*)(dta+24),(word*)(dta+22));
    if (stddebug) fprintf(stddebug,"FINDFILE: findfile, found %s\n",dta+30);
    CLC;
    return 1;
  }
  free(DTA->path);
  free(DTA->mask);
  closedir(DTA->dir);
  STC;
  AX=DERRNO(DNOMOREFILES);
  if (stddebug) fprintf(stddebug,"FINDFILE: no more files\n");
  return 1;
}


static __inline__ int int21_emu_findfirst ( void )
{
  char *dta=DOS_PB(dta_seg,dta_ofs),*p;
  int r=path_d2u(DOS_PB(DS,DX),DIR_ACCESS);
  if (r) {
    if (stddebug) fprintf(stddebug,"FINDFIRST: path conv failed\n");
    AX=DERRNO(r);
    STC;
    return 1;
  }
  p=strrchr(path,'/');
  *(p++)='\0';
  if (stddebug) fprintf(stddebug,"FINDFIRST: dir=\"%s\" mask=\"%s\" attrmask(CX)=%04Xh\n",path,p,CX);
  if ((DTA->dir=opendir(*path?path:"/"))==NULL) {
    AX=DERRNO(errno_u2d(errno));
    STC;
    return 1;
  }
  DTA->root=dir_isdosroot(last_drive,*path?path:"/");
  if (DTA->root&&stddebug)  fprintf(stddebug,"ROOT DIRECTORY\n");
  DTA->path=*path?dosrun_strdup(path):dosrun_strdup("/");
  DTA->mask=dosrun_strdup(p);
  DTA->attr=CX;
  return em_findfile(dta);
}



static __inline__ int int21_emu_findnext ( void )
{
  return em_findfile(DOS_PB(dta_seg,dta_ofs));
}
