/*  DOSRUN, main.c                                         *
 *  (C)1999 LGB (Gabor Lenart), lgb@vlug.vein.hu           *
 *  Free software under the terms of GNU/GPL v2 or later   */


#include <stdio.h>
#include <fcntl.h>
#include <errno.h>
#include <signal.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "config_dosrun.h"
#include "dosrun.h"

#include "memory.h"
#include "derrno.h"
#include "loader.h"
#include "xms.h"

#include "main.h"
#include "options.h"
#include "filename.h"
#include "screen.h"


FILE *stddebug=NULL;
struct vm86plus_struct vm;
static int in_vm86=0;
int exitcode=0;
extern char **environ;
word oldcs,oldip;

int derrno;
char *env=NULL;


#include "main_common.c"

/* yes, ugly */
static __inline__ int dosrun_vm86 ( int mode )
{
  int r;
  in_vm86=1;
  __asm__ volatile (
  "int $0x80"
  : "=a" (r)
  : "0" (166), "b" (mode), "c" (&vm));
  in_vm86=0;
  return r;
}


#include "main_emu.c"

static int in_sig_handler=0;


static void sighandler ( int signum )
{
	fprintf(stderr,"sighandler: got signal %d.\n",signum);
	if (in_sig_handler) {
	  fprintf(stderr,"sighandler: Signal in sighandler !!!!!!\n");
		exit(1);
	}
  in_sig_handler=1;
	fprintf(stderr,"sighandlder: starting shutting down dosrun ...\n");
  screen_uninit();
  fprintf(stderr,"Signal %d received in %s mode.\n",signum,in_vm86?"vm86":"normal");
  dump(1);
  dosmem_dump();
  exit(1);
}


__inline__ static void init_sighandler ( void )
{
  int a;
return;
	for (a=0;a<32;a++) signal(a,sighandler);
}


__inline__ static int execute ( int argc , char ** argv , char * env )
{
  int addr;
  addr=load_prg(argv[0],env);
  if (addr<=0) {
    fprintf(stderr,"Error loading executable \"%s\".\n",argv[0]);
    exit(1);
  }
  free(env);
  pass_commandline(addr,argc-1,argv+1);
  screen_init();
  if (stddebug) fprintf(stddebug,"Starting executing program ...\n");
  while (emu());
  dosmem_dump();    /* see MCB list at the end for debugging */
  return exitcode;
}



/* hmmm. Hard to understand ? :)
   dosrun must allow dos apps to chain interrupts.
   we set up a table in "rom" area with az INT instrucion for each
   interrupt (interrupt number does not matter, so use 0xCD because
   int opcode is 0xcd as well). interrupt handler (main_emu.c) checks
   if interrupt occured in THIS table. If it's process DOS interrupt
   if no, just emulate an interrupt call, taken from interrupt table
   located at memory location of zero */
__inline__ static void init_int ( void )
{
  int a;
  word *w=NULL;
  for (a=0;a<256;a++) {
    *(w++)=EMU_INT_OFS+a;
    *(w++)=EMU_INT_SEG;
  }
  memset(DOS_PB(EMU_INT_SEG,EMU_INT_OFS),0xCD,0x101);
  memset(&vm.int_revectored,255,32);
  memset(&vm.int21_revectored,255,32);
}




int main ( int argc , char **argv )
{
  int r;
  if (!getuid()||getuid()!=geteuid()) {
    fprintf(stderr,"DO NOT run dosrun as root or with (neither with not root) setuid bit !!\n");
    exit(1);
  }
  if (argc<2) {
    options_help(argv[0]);
    exit(1);
  }
  /* option parser may override compile default values so we adjust all settings here */
  dosdrive_init();
  dosmem_init(CONFIG_DOSMEM_START,CONFIG_DOSMEM_SIZE<<10);
#ifdef CONFIG_XMS
  xms_params(XMS_SIZE<<10,XMS_HANDLES,HMA_MIN); /* set compile default values */
#endif  
  bzero(&vm,sizeof(vm));
  vm.cpu_type=CONFIG_CPU_TYPE;
  config.fullscreen=0;
	config.autofullscreen=1;
	r=options_parser(argc,argv);
  if (r<0) exit(-r);
  if (dosrun_vm86(VM86_PLUS_INSTALL_CHECK)) {
    fprintf(stderr,"Your kernel does not have support for vm86plus(). Upgrade!\n");
    exit(1);
  }
#ifdef CONFIG_XMS
  if (xms_init()) {
    fprintf(stderr,"Cannot initialize XMS emulation.\n");
    exit(1);
  }
#endif
  init_int();
  *DOS_PB(INT21_38_CAPITALPROC_SEG,INT21_38_CAPITALPROC_OFS)=0xCB;
  *DOS_PB(0xF000,0xFFFE)=0xFC;  /* 0xFC=PC/AT */
  EFLAGS=IF_MASK|IOPL_MASK;
  init_sighandler();
  dosmem_romprotect(); /* protecting "ROM" area is now ON */
  env_dump(env);
  screen_terminal_test();
  r=execute(argc-r,argv+r,env);
  screen_uninit();
  return r;
}
