/* _RAW_ screen handling for DOSRUN project
 * (C)1999 LGB of DC, Gabor Lenart, lgb@vlug.vein.hu
 *  GNU/GPL code
 */

#include <stdio.h>
#include "config_dosrun.h"
#include "dosrun.h"
#include "options.h"
#include <sys/ioctl.h>
#include <term.h>
#include "translate.h"

void dosrun_fatal ( char * );



#ifdef CONFIG_FULL_SCREEN
#include <curses.h>
#include <pthread.h>
#include <unistd.h>
static int CGA_colors[] = { 0,4,2,6,1,5,3,7 };
static attr_t CGA_attr[256];
pthread_t refresh_thread;
int scrxsize=80;
int scrysize=25;
int thread_enabled;
int thread_force=0;
extern int curx,cury;
#endif

#include "screen.h"

__inline__ chtype screen_translate_char ( byte c );

void screen_terminal_test ( void )
{
  struct termios term;
  config.termout=!ioctl(1,TCGETS,&term);
  config.termin =!ioctl(0,TCGETS,&term);
  if (!config.termout) {
    config.fullscreen=0;
    config.autofullscreen=0;
  }
}


__inline__ static void screen_init_ncurses ( void )
{
  int a;
  initscr();
  cbreak();
//  nocbreak();
//  noecho();
  echo();
	keypad(stdscr,TRUE);
  start_color();
  for (a=0;a<COLOR_PAIRS;a++)
    init_pair(a,CGA_colors[a%COLORS],CGA_colors[a/COLORS]);
  for (a=0;a<256;a++) {
    int attr=COLOR_PAIR(((a&0x70)>>1)|(a&7));
    if (a&128) attr|=A_BLINK;
    if (a&8) attr|=A_BOLD;
    CGA_attr[a]=attr;
  }
	config.fullscreen=1;
}


static void screen_refresh ( void )
{
  if (vm.screen_bitmap&0x01000000) {
    int x,y;
    byte *p=(byte*)0xB8000;
    vm.screen_bitmap=0;
    for (y=0;y<scrysize;y++)
      for (x=0;x<scrxsize;x++) {
        attrset(CGA_attr[p[1]]);
        mvaddch(y,x,screen_translate_char(*p));
	      p+=2;
      }
    attrset(A_NORMAL);
		move(cury,curx);
    refresh();
    if (stddebug) fprintf(stddebug,"THREAD: refresh!\n");
  }
}



static void refresh_thread_func ( void )
{
  if (config.autofullscreen) {
    while (!thread_force) {
      if (!thread_enabled) {
        if (stddebug) fprintf(stddebug,"THREAD: exit\n");
        pthread_exit(NULL);
      }
      if (vm.screen_bitmap) thread_force=1;
      usleep(10000);
    }
    if (stddebug&&thread_enabled) fprintf(stddebug,"THREAD: direct video ram write detected.\nTHREAD: swithcing into full-screen mode.\n");    
    screen_init_ncurses();
  }
  while (thread_enabled) {
    screen_refresh();
    usleep(50000);
  }
  if (stddebug) fprintf(stddebug,"THREAD: exit\n");
  pthread_exit(NULL);
}


__inline__ static void screen_init_base ( void )
{
  int a;
  vm.flags=VM86_SCREEN_BITMAP;
  for (a=0;a<16348;a++) *((word*)0xB8000+a)=0x0720;
  vm.screen_bitmap=0;
}




void screen_init ( void )
{
  screen_init_base();
  if (config.autofullscreen||config.fullscreen) {
    if (config.fullscreen) screen_init_ncurses();
    thread_enabled=1;
    if (pthread_create(&refresh_thread,NULL,refresh_thread_func,NULL))  dosrun_fatal("Cannot create thread for screen refreshing");
    if (stddebug) fprintf(stddebug,"THREAD: screen refresh thread created.\n");
  }
}


void screen_uninit ( void )
{
  fprintf(stderr,"SCREEN: uninit\n");
	fflush(stderr);
  if (config.fullscreen||config.autofullscreen) {
    thread_enabled=0;
		if (stddebug) fprintf(stddebug,"waiting for shutdown of screen thread ... ");
    pthread_join(refresh_thread,NULL);
		if (stddebug) fprintf(stddebug,"DONE\n");
  }
  if (config.fullscreen) endwin();
}

